use std::{cell::OnceCell, env, str::FromStr, sync::Arc};

use rand::{distributions::Alphanumeric, thread_rng, Rng};
use regex::{Captures, Regex};
use sqlx::{postgres::{PgConnectOptions, PgPoolOptions}, Connection, Executor, PgConnection, PgPool};


pub struct DbTestContext {
    //pub app: Router,
    pub db_pool: PgPool,
}

#[proc_macro_attribute]
pub fn db_test(_: TokenStream, item: TokenStream) -> TokenStream {
    let input = parse_macro_input!(item as ItemFn);
    let test_name = input.sig.ident.clone();
    let test_arguments = input.sig.inputs;
    let test_block = input.block;
    let inner_test_name = syn::Ident::new(
        format!("inner_{}", test_name).as_str(),
        input.sig.ident.span(),
    );

    let setup = quote! {
        let context = abc_web::test_helpers::setup().await;
    };

    let teardown = quote! {
        abc_web::test_helpers::teardown(context).await;
    };

    let output = quote!(
        #[::tokio::test]
        async fn #test_name() {
            #setup
            async fn #inner_test_name(#test_arguments) #test_block
            #inner_test_name(&context).await;
            #teardown
        }
    );

    TokenStream::from(output)
}

#[allow(unused)]
pub async fn setup() -> DbTestContext {
    //let init_config: OnceCell<Config> = OnceCell::new();
    //let config = init_config.get_or_init(|| load_config(&Environment::Test).unwrap());

    let test_db_pool = setup_db(&env::var("DATABASE_URL").unwrap()).await;

    //let app = init_routes(AppState {
    //    db_pool: test_db_pool.clone(),
    //});

    DbTestContext {
        //app,
        db_pool: test_db_pool,
    }
}

#[allow(unused)]
pub async fn teardown(context: DbTestContext) {
    //drop(context.app);

    teardown_db(context.db_pool);
}

#[allow(unused)]
pub async fn setup_db(url: &str) -> PgPool {
    let test_db_config = prepare_db(url).await;

    let pool = PgPoolOptions::new()
        .connect(url)
        .await
        .unwrap();

    pool
}

/// Drops a dedicated database for a test case.
///
/// This function is automatically called by the [`abc-macros::db_test`] macro. It ensures test-specific database are cleaned up after each test run so we don't end up with large numbers of unused databases.
pub async fn teardown_db(pool: PgPool) {
    let cloned_pool = pool.clone();
    //let mut connect_options = pool.connect_options();
    //let db_config = Arc::make_mut(&mut connect_options);
    let db_config = cloned_pool.connect_options();

    drop(pool);

    let root_db_config = db_config.clone().database("postgres");
    let mut connection: PgConnection = Connection::connect_with(&root_db_config).await.unwrap();

    let test_db_name = db_config.get_database().unwrap();

    let query = format!("DROP DATABASE IF EXISTS {}", test_db_name);
    connection.execute(query.as_str()).await.unwrap();
}

async fn prepare_db(url: &str) -> String {
    let db_config = PgConnectOptions::from_str(url).expect("Invalid DATABASE_URL!");
    let db_name = db_config.get_database().unwrap();

    let root_db_config = db_config.clone().database("postgres");
    let mut connection: PgConnection = Connection::connect_with(&root_db_config).await.unwrap();

    let test_db_name = build_test_db_name(db_name);

    let query = format!("CREATE DATABASE {} TEMPLATE {}", test_db_name, db_name);
    connection.execute(query.as_str()).await.unwrap();

    let regex = Regex::new(r"(.+)\/(.+$)").unwrap();
    let test_db_url = regex.replace(url, |caps: &Captures| {
        format!("{}/{}", &caps[1], test_db_name)
    });

    test_db_url.to_string()
}

fn build_test_db_name(base_name: &str) -> String {
    let test_db_suffix: String = thread_rng()
        .sample_iter(&Alphanumeric)
        .take(30)
        .map(char::from)
        .collect();
    format!("{}_{}", base_name, test_db_suffix).to_lowercase()
}
